#ifndef __FLOATS_H_INCLUDED__
#define __FLOATS_H_INCLUDED__

///////////////////////////////////////////////////////////////////////////////
#define MATH_EPSILON 0.00001f

class float2
{
 public:
 
	 float x, y;

	 float2() {}

	 float2(float _x, float _y)
	 {
		 x = _x;
		 y = _y;

		 return;
	 }
};

inline bool operator == (const float2 &a, const float2 &b)
{
 if (fabs(a.x - b.x) > MATH_EPSILON) return false;
 if (fabs(a.y - b.y) > MATH_EPSILON) return false;
 return true;
}

class float3
{
 public:
	
	 float x, y, z;

	 float3() {}

	 float3(float _x, float _y, float _z)
	 {
		 x = _x;
		 y = _y;
		 z = _z;

		 return;
	 }

	 float3(const float2 &v)
	 {
		 x = v.x;
		 y = v.y;
		 z = 0.0;

		 return;
	 }

	 float2 xy() const
	 {
		 return float2(x, y);
	 }

	 float3 operator - (const float3 &oth) const
	 {
		 return float3(x - oth.x, 
					   y - oth.y, 
					   z - oth.z);
	 }

	 float3 operator * (const float &oth) const
	 {
		 return float3(x * oth, y * oth, z * oth);
	 }
};

inline float3 normalize(const float3 &v)
{
 float l = (v.x*v.x)+(v.y*v.y)+(v.z*v.z);
 if (l < MATH_EPSILON)
	 return float3(0.0, 0.0, 0.0);

 return v * (1.0f / sqrtf(l));
}

inline float3 cross(const float3 &a, const float3 &b)
{
 return float3(a.y * b.z - a.z * b.y,
			   a.z * b.x - a.x * b.z,
			   a.x * b.y - a.y * b.x);
}

inline float dot(const float3 &a, const float3 &b)
{
 return (a.x * b.x) + (a.y * b.y) + (a.z * b.z);
}

inline bool operator == (const float3 &a, const float3 &b)
{
 if (fabs(a.x - b.x) > MATH_EPSILON) return false;
 if (fabs(a.y - b.y) > MATH_EPSILON) return false;
 if (fabs(a.z - b.z) > MATH_EPSILON) return false;
 return true;
}

class float4
{
 public:
 
	 float x, y, z, w;

	 float4() {}

	 float4(float _x, float _y, float _z, float _w)
	 {
		 x = _x;
		 y = _y;
		 z = _z;
		 w = _w;

		 return;
	 }

	 float4(const float2 &v)
	 {
		 x = v.x;
		 y = v.y;
		 z = 0.0;
		 w = 0.0;

		 return;
	 }

	 float4(const float3 &v)
	 {
		 x = v.x;
		 y = v.y;
		 z = v.z;
		 w = 0.0;

		 return;
	 }

	 float2 xy() const
	 {
		 return float2(x, y);
	 }

	 float3 xyz() const
	 {
		 return float3(x, y, z);
	 }
};

inline bool operator == (const float4 &a, const float4 &b)
{
 if (fabs(a.x - b.x) > MATH_EPSILON) return false;
 if (fabs(a.y - b.y) > MATH_EPSILON) return false;
 if (fabs(a.z - b.z) > MATH_EPSILON) return false;
 if (fabs(a.w - b.w) > MATH_EPSILON) return false;
 return true;
}
///////////////////////////////////////////////////////////////////////////////
#endif // __FLOATS_H_INCLUDED__
