#define  _CRT_SECURE_NO_WARNINGS
#define  GLEW_STATIC
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <vector>
#include <Windows.h>
#include "GL/glew.h"
#include "GL/glut.h"
#include "Floats.h"
#include "Matrix.h"

typedef struct _c_OBJ_VERTEX
{
 float3 pos;
 float2 uv;
 float3 normal;
}c_OBJ_VERTEX;

static std::vector<c_OBJ_VERTEX> gObjectVertexes;
static std::vector<WORD>		 gObjectFaces;
static GLuint                    gVertexBuffer;
static GLuint					 gIndexBuffer;
static GLuint					 gVertexProgram;
static GLuint					 gFragmentProgram;
static GLuint					 gLinkedProgram;
static GLuint					 gParam_VPMatrix;
static GLuint					 gParam_WMatrix;

//////////////////////////////////////////////////////////////////////
void XPPrint(const char *format, ...)
{
 if (format == NULL)
	 return;

 va_list arglist;
 static char buffer[8192];

 va_start(arglist, format);
 vsprintf_s(buffer, sizeof(buffer)/sizeof(buffer[0]), format, arglist);
 va_end(arglist);	 

 ::OutputDebugStringA(buffer);
 ::OutputDebugStringA("\n");

 printf("%s\n", buffer);

 return;
}
//////////////////////////////////////////////////////////////////////
void GLERR()
{
 GLenum err = glGetError();
 if (err == GL_NO_ERROR)
	 return;

 switch (err)
 {
	case GL_INVALID_ENUM:
		XPPrint("GL_INVALID_ENUM");
	break;
	case GL_INVALID_VALUE:
		XPPrint("GL_INVALID_VALUE");
	break;
	case GL_INVALID_OPERATION:
		XPPrint("GL_INVALID_OPERATION");
	break;
	case GL_STACK_OVERFLOW:
		XPPrint("GL_STACK_OVERFLOW");
	break;
	case GL_STACK_UNDERFLOW:
		XPPrint("GL_STACK_UNDERFLOW");
	break;
	case GL_OUT_OF_MEMORY:
		XPPrint("GL_OUT_OF_MEMORY");
	break;
	case GL_INVALID_FRAMEBUFFER_OPERATION:
		XPPrint("GL_INVALID_FRAMEBUFFER_OPERATION_OES");
	break;
	default:
		XPPrint("GL_UNKNOWN (%X)", err);
 }

#ifdef _DEBUG
 _asm { int 3 };
#endif

 return;
}
//////////////////////////////////////////////////////////////////////
void LoadOBJ(const char *filename)
{
 FILE *f = fopen(filename, "rb");
 if (f == NULL)
 {
	 XPPrint("Cannot open file [%s]", filename);
	 exit(0);
 }

 std::vector<float3> Positions;
 std::vector<float3> Normals;
 std::vector<float2> UVs;

 char buffer[1024] = { 0 };
 while (fgets(buffer, 1024, f))
 {
	 if (buffer[0] == 'v' && buffer[1] == ' ')
	 {
		 // Vertexes
		 float3 vert;
		 sscanf(&buffer[2], "%f %f %f", &vert.x, &vert.y, &vert.z);
		 Positions.push_back(vert);
		 continue;
	 }

	 if (buffer[0] == 'v' && buffer[1] == 't' && buffer[2] == ' ')
	 {
		 // UVs
		 float2 uv;
		 sscanf(&buffer[3], "%f %f", &uv.x, &uv.y);
		 UVs.push_back(uv);
	 }

	 if (buffer[0] == 'v' && buffer[1] == 'n' && buffer[2] == ' ')
	 {
		 // Normals
		 float3 n;
		 sscanf(&buffer[3], "%f %f %f", &n.x, &n.y, &n.z);
		 Normals.push_back(n);
	 }

	 if (buffer[0] == 'f' && buffer[1] == ' ')
	 {
		 int Vrt[3], Nrm[3], Uv[3];
		 sscanf(&buffer[2], "%i/%i/%i %i/%i/%i %i/%i/%i", &Vrt[0], &Uv[0], &Nrm[0],
														  &Vrt[1], &Uv[1], &Nrm[1],
														  &Vrt[2], &Uv[2], &Nrm[2]);

		 c_OBJ_VERTEX Vtx;
		 for (int i = 0; i < 3; i++)
		 {
			 Vtx.pos    = Positions[Vrt[i]-1];
			 Vtx.normal = Normals  [Nrm[i]-1];
			 Vtx.uv     = UVs      [Uv [i]-1];

			 BOOL Found = FALSE;

			 for (size_t j = 0; j < gObjectVertexes.size(); j++)
			 {
				 if (Vtx.pos    == gObjectVertexes[j].pos    &&
					 Vtx.normal == gObjectVertexes[j].normal &&
					 Vtx.uv     == gObjectVertexes[j].uv      )
				 {
					 gObjectFaces.push_back((WORD)j);
					 Found = TRUE;
					 break;
				 }
			 }

			 if (Found == FALSE)
			 {
				 gObjectFaces.push_back((WORD)gObjectVertexes.size());
				 gObjectVertexes.push_back(Vtx);
			 }
		 }
	 }
 }

 fclose(f);

 glGenBuffers(1, &gVertexBuffer);
 glGenBuffers(1, &gIndexBuffer);
 GLERR();

 glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, gIndexBuffer);
 GLERR();
 glBufferData(GL_ELEMENT_ARRAY_BUFFER, gObjectFaces.size()*sizeof(WORD), &gObjectFaces[0], GL_STATIC_DRAW);
 GLERR();
 glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
 GLERR();

 glBindBuffer(GL_ARRAY_BUFFER, gVertexBuffer);
 GLERR();
 glBufferData(GL_ARRAY_BUFFER, gObjectVertexes.size()*sizeof(c_OBJ_VERTEX), &gObjectVertexes[0], GL_STATIC_DRAW);
 GLERR();
 glBindBuffer(GL_ARRAY_BUFFER, 0);
 GLERR();

 return;
}
//////////////////////////////////////////////////////////////////////
GLenum CreateShader(const char *filename, GLenum type)
{
 FILE *f = fopen(filename, "rb");
 if (f == NULL)
 {
	 XPPrint("Cannot open file [%s]", filename);
	 exit(0);
 }

 fseek(f, 0, SEEK_END);
 DWORD s = ftell(f);
 fseek(f, 0, SEEK_SET);

 char *buffer = (char*)malloc(s+1);
 fread(buffer, s, 1, f);
 fclose(f);
 buffer[s] = 0;

 GLenum Shader = glCreateShaderObjectARB(type);
 GLERR();

 glShaderSourceARB(Shader, 1, (const GLcharARB **)&buffer, NULL);
 GLERR();

 free(buffer);

 glCompileShader(Shader);
 GLERR();

 int Compiled = 0;
 glGetObjectParameterivARB(Shader, GL_OBJECT_COMPILE_STATUS_ARB, &Compiled);
 GLERR();
 
 if (!Compiled) 
 { 
	 XPPrint("There was error during compilation !");

	 int BLen = 0;	 
	 int SLen = 0;	

	 glGetObjectParameterivARB(Shader, GL_OBJECT_INFO_LOG_LENGTH_ARB , &BLen);
	 GLERR();

	 if (BLen > 1)
	 {
		 char *LogData = (char *)malloc(BLen+1);
	
		 glGetInfoLogARB(Shader, BLen, &SLen, LogData);
		 GLERR();

		 XPPrint("Compilation error log: \n%s\n", LogData);

         free(LogData);   
	}

	glDeleteObjectARB(Shader);
	return 0;
 }

 return Shader;
}
//////////////////////////////////////////////////////////////////////
void Initialize()
{
	GLERR();
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_LESS);
    glClearColor(0.2, 0.2, 0.2, 1.0);
	glewInit();

	//
	// Load object
	//
	GLERR();
	LoadOBJ("torus.obj");
	GLERR();

	//
	// Load programs
	//
	gVertexProgram   = CreateShader("VertexProgram.glsl", GL_VERTEX_SHADER_ARB);
	gFragmentProgram = CreateShader("FragmentProgram.glsl", GL_FRAGMENT_SHADER_ARB);

	//
	// Link programs
	//
	gLinkedProgram = glCreateProgramObjectARB();
	GLERR();

	glAttachObjectARB(gLinkedProgram, gVertexProgram);
	GLERR();

	glAttachObjectARB(gLinkedProgram, gFragmentProgram);
	GLERR();

	glLinkProgramARB(gLinkedProgram);
	GLERR();

	int Linked = 0;
	glGetObjectParameterivARB(gLinkedProgram, GL_OBJECT_LINK_STATUS_ARB, &Linked);
	GLERR();
	 
	if (!Linked)
	{
		XPPrint("There was error during linking !");
	    
		int BLen = 0;	 
	    int SLen = 0;	

	    glGetObjectParameterivARB(gLinkedProgram, GL_OBJECT_INFO_LOG_LENGTH_ARB , &BLen);
		GLERR();

		if (BLen > 1)
		{
			char *LogData = (char *)malloc(BLen+1);
		
			 glGetInfoLogARB(gLinkedProgram, BLen, &SLen, LogData);
			 GLERR();

			 XPPrint("Linking error log: \n%s\n", LogData);
			 free(LogData);   
		}

		glDeleteObjectARB(gLinkedProgram);
		GLERR();

		gLinkedProgram = 0;
	}
	
	if (gLinkedProgram)
	{
		gParam_VPMatrix = glGetUniformLocationARB(gLinkedProgram, "VPMatrix");
		gParam_WMatrix  = glGetUniformLocationARB(gLinkedProgram, "WMatrix");
	}

	GLERR();

	return;
}
//////////////////////////////////////////////////////////////////////
void Display() 
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	GLERR();

	glUseProgramObjectARB(gLinkedProgram);
	GLERR();

	matrix Camera; Camera.MakeLookAtMatrix(50.0, 0.0, 0.0,
											0.0, 0.0, 0.0,
											0.0, 1.0, 0.0);

	matrix Projection; Projection.MakeProjectionMatrix(45.0, 1.0, 1.0, 1024.0);

	matrix WVP = Projection * Camera;

	glUniform4fvARB(gParam_VPMatrix, 4, WVP.m.v);
	GLERR();

	matrix WMat;
		   WMat.MakeRotation((float)GetTickCount()/1000.0f, 1.0f, 1.0f, 1.0f);

	glUniform4fvARB(gParam_WMatrix, 4, WMat.m.v);
	GLERR();

	//
	// Set clipping plane (gl_ClipVertex is in WORLD space, so out clip plane in WORLD space to)
	//
	double cf[4] = { 0.0, 0.0, 1.0, 0.0 };

	glClipPlane(GL_CLIP_PLANE0, cf);
	GLERR();

	glEnable(GL_CLIP_PLANE0);
	GLERR();

	//
	// Bind buffers
	//
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, gIndexBuffer);
	GLERR();

	glBindBuffer(GL_ARRAY_BUFFER, gVertexBuffer);
	GLERR();

	glEnableClientState  (GL_VERTEX_ARRAY);
	glVertexPointer      (3, GL_FLOAT, sizeof(c_OBJ_VERTEX), offsetof(c_OBJ_VERTEX, pos.x));
	GLERR();

	glClientActiveTexture(GL_TEXTURE1);
    glEnableClientState  (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer    (3, GL_FLOAT, sizeof(c_OBJ_VERTEX), offsetof(c_OBJ_VERTEX, normal.x));
	GLERR();

	glClientActiveTexture(GL_TEXTURE0);
    glEnableClientState  (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer    (2, GL_FLOAT, sizeof(c_OBJ_VERTEX), offsetof(c_OBJ_VERTEX, uv.x));
	GLERR();

	//
	// Validate program
	//
	static int ProgramValidated = FALSE;

	if (ProgramValidated == FALSE)
	{
		ProgramValidated = TRUE;

		glValidateProgramARB(gLinkedProgram);
		GLERR();

		int Valid = 0;
		glGetObjectParameterivARB(gLinkedProgram, GL_OBJECT_VALIDATE_STATUS_ARB, &Valid);
		GLERR();

		if (!Valid)
		{
			int BLen = 0;	 
			int SLen = 0;	

			XPPrint("Error while validating program !");

			glGetObjectParameterivARB(gLinkedProgram, GL_OBJECT_INFO_LOG_LENGTH_ARB , &BLen);
			GLERR();

			if (BLen > 1)
			{
				 char *LogData = (char *)malloc(BLen+1);
			
				 glGetInfoLogARB(gLinkedProgram, BLen, &SLen, LogData);
				 GLERR();

				 XPPrint("Validation error log: \n%s\n", LogData);
				 free(LogData);  
			}
		}
		else
		{
			int BLen = 0;	 
			int SLen = 0;	

			XPPrint("Program validated successfully !");

			glGetObjectParameterivARB(gLinkedProgram, GL_OBJECT_INFO_LOG_LENGTH_ARB , &BLen);
			GLERR();

			if (BLen > 1)
			{
				 char *LogData = (char *)malloc(BLen+1);
			
				 glGetInfoLogARB(gLinkedProgram, BLen, &SLen, LogData);
				 GLERR();

				 XPPrint("Validation warnings log: \n%s\n", LogData);
				 free(LogData);  
			}
		}
	}

	//
	// Draw primitive
	//
	glDrawRangeElements(GL_TRIANGLES, 
						0, 
						gObjectVertexes.size()-1,
						gObjectFaces.size(),
						GL_UNSIGNED_SHORT,
						NULL);
	GLERR();

	//
	// Restore states
	//
	glDisable(GL_CLIP_PLANE0);
	GLERR();

	glDisableClientState(GL_VERTEX_ARRAY);
	GLERR();

	glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	GLERR();

	glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	GLERR();

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	GLERR();

	glBindBuffer(GL_ARRAY_BUFFER, 0);
	GLERR();

	glUseProgramObjectARB(0);
	GLERR();

	glutSwapBuffers();
	GLERR();

	return;
}
//////////////////////////////////////////////////////////////////////
void Idle() 
{
    glutPostRedisplay();
	GLERR();
}
//////////////////////////////////////////////////////////////////////
void Key(unsigned char k, int x, int y) 
{
    k = tolower(k);

    switch(k) 
	{
        case 27:
        case 'q':
			exit(0);
        break;
    }
    
	glutPostRedisplay();
	GLERR();
}
//////////////////////////////////////////////////////////////////////
void Resize(int w, int h) 
{
	glViewport(0, 0, w, h);
	GLERR();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	GLERR();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	GLERR();

	return;
}
//////////////////////////////////////////////////////////////////////
int main(int argc, char **argv) 
{
	glutInit(&argc, argv);
	glutInitWindowSize(512, 512);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
	glutCreateWindow("Just a test");
	GLERR();

	Initialize();

	glutDisplayFunc(Display);
    glutIdleFunc(Idle);
    glutKeyboardFunc(Key);
    glutReshapeFunc(Resize);

	GLERR();
	glutMainLoop();
	GLERR();

	return 0;
}
//////////////////////////////////////////////////////////////////////
